import click
from packyak.cli.cli import cli
from packyak.util.emr import EMR
import asyncio
import os


@cli.command()
@click.option(
    "--cluster-id",
    required=True,
    type=str,
    help="The unique identifier of the EMR cluster.",
)
@click.option(
    "--instance-id",
    required=True,
    type=str,
    help="The unique identifier of the EMR instance.",
)
@click.option(
    "--profile",
    required=False,
    type=str,
    default=None,
    help="The AWS profile to use for the session.",
)
@click.argument("log-files", nargs=-1, type=str)
@click.option(
    "--get",
    is_flag=True,
    help="Retrieve and display the content of the log files.",
)
def logs(
    cluster_id: str,
    instance_id: str,
    profile: str | None,
    log_files: list[str] = [],
    get: bool = False,
):
    """
    Lists all log files for a given EMR cluster and instance ID.
    """
    if profile is not None:
        os.environ["AWS_PROFILE"] = profile
    emr = EMR()

    async def list_and_print_logs():
        logs = await emr.list_logs(cluster_id, instance_id)
        import fnmatch

        filtered_logs = logs
        if log_files:
            filtered_logs = []
            for log_file_pattern in log_files:
                filtered_logs.extend(fnmatch.filter(logs, log_file_pattern))

        if get:
            log_text = await asyncio.gather(
                *[emr.get_log_text(cluster_id, log) for log in filtered_logs]
            )
            for text in log_text:
                print(text)
        else:
            for log in filtered_logs:
                print(log)

    asyncio.run(list_and_print_logs())



from packyak.cli.cli import cli
from packyak.cli.list import list
from packyak.cli.materialize import materialize
from packyak.cli.new import new
from packyak.cli.run import run
from packyak.cli.ssh import ssh
from packyak.cli.synth import synth
from packyak.cli.logs import logs
from packyak.cli.instances import instances


__all__ = [
    "cli",
    "new",
    "ssh",
    "synth",
    "run",
    "materialize",
    "list",
    "logs",
    "instances",
]

cli()



import asyncio
from typing import Any
import click
import subprocess
import time
import boto3
import os
import re
import collections
import aiofiles

from packyak.cli.cli import cli
from packyak.util.emr import EMR

from aiobotocore.session import get_session

session = get_session()


@cli.command()
@click.argument(
    "instance-id",
    type=str,
    # prompt="What is the ID of the EC2 Instance or EMR Cluster you wish to tunnel to?",
    # help="The EC2 Instance ID of the instance you wish to tunnel to",
)
@click.option(
    "--ssh-key",
    default="~/.ssh/id_rsa",
    help="Path to the the SSH public key to copy to the remote host. Default: ~/.ssh/id_rsa.pub",
)
@click.option(
    "-L",
    "port_forwards",
    # type=(str | list[str]),
    multiple=True,
    help="Ports to forward. Default: 9000:localhost:22",
)
@click.option("-v", "--verbose", type=str, is_flag=True)
@click.option(
    "--profile", type=str, help="AWS CLI profile to use when authenticating to SSM"
)
@click.option("--user", type=str, help="POSIX username to log in as. Defaults to root.")
def ssh(
    instance_id: str,
    ssh_key: str = "~/.ssh/id_rsa",
    # todo: what port forwarding to i need for VS Code Remote SSH
    port_forwards: list[str] = ["9001:localhost:22"],
    verbose: bool = False,
    profile: str | None = None,
    user: str | None = None,
):
    """
    Establishes a secure tunnel to an EC2 instance.

    instance-id: The EC2 Instance ID of the instance you wish to tunnel to.

    --ssh-public-key: Path to the SSH public key to copy to the remote host. Default: ~/.ssh/id_rsa.pub

    -L: Ports to forward. Default: 9001:localhost:22
    """

    if profile is not None:
        os.environ["AWS_PROFILE"] = profile
    emr = EMR()

    def log(message: str):
        if verbose:
            print(message)

    async def ssh(instance_id: str, user: str | None):
        if instance_id.startswith("j"):
            log(
                f"Instance ID {instance_id} looks like an EMR cluster ID. Resolving the Primary node."
            )
            instance_id = await emr.get_primary_node_instance_id(cluster_id=instance_id)
            log(f"Resolved primary node: {instance_id}")

        if user is None:
            async with session.create_client("sts") as sts_client:
                caller_identity = await sts_client.get_caller_identity()

            user_id = caller_identity.get("UserId", "")
            user = user_id.split(":")[1] if ":" in user_id else "root"
        log(f"Logging in as {user}")

        log(f"{time.ctime()} sm-connect-ssh-proxy: Connecting to: {instance_id}")
        # print(f"{time.ctime()} sm-connect-ssh-proxy: Extra args: {extra_ssh_args}")

        async with session.create_client("ssm") as ssm_client:
            instance_info = await ssm_client.describe_instance_information(
                Filters=[{"Key": "InstanceIds", "Values": [instance_id]}]
            )
            if not instance_info["InstanceInformationList"]:
                log("No instance information found.")
                instance_status = "Offline"
            else:
                instance_status = instance_info["InstanceInformationList"][0][
                    "PingStatus"
                ]
            log(f"Instance status: {instance_status}")

            if instance_status != "Online":
                log("Error: Instance is offline.")
                return

            async with aiofiles.open(os.path.expanduser(f"{ssh_key}.pub"), "r") as file:
                ssh_pub_key = (await file.read()).strip()

            current_region = boto3.session.Session().region_name
            log(f"Will use AWS Region: {current_region}")

            aws_cli_version = subprocess.check_output(
                ["aws", "--version"], stderr=subprocess.STDOUT
            ).decode()
            log(f"AWS CLI version (should be v2): {aws_cli_version}")

            ssh_dir = "/root/.ssh/" if user == "root" else f"/home/{user}/.ssh/"

            commands = [
                # f'echo "{ssh_pub_key}" > /etc/ssh/authorized_keys',
                f"mkdir -p {ssh_dir}",
                f'echo "{ssh_pub_key}" > {ssh_dir}/authorized_keys',
            ]
            log(f"Sending command to instance: {commands}")
            send_command_response = await ssm_client.send_command(
                InstanceIds=[instance_id],
                DocumentName="AWS-RunShellScript",
                Comment="Copy public key for SSH helper",
                Parameters={"commands": commands},
                TimeoutSeconds=30,
            )
            command_id = send_command_response["Command"]["CommandId"]
            log(f"Got command ID: {command_id}")

            # time.sleep(5)  # Wait a bit to prevent InvocationDoesNotExist error

            for _ in range(15):
                try:
                    output = await ssm_client.get_command_invocation(
                        CommandId=command_id, InstanceId=instance_id
                    )
                    log(f"Command status: {output['Status']}")
                    if output["Status"] not in ["Pending", "InProgress"]:
                        log(
                            f"Command output: {output.get('StandardOutputContent', '')}"
                        )
                        if output.get("StandardErrorContent"):
                            log(f"Command error: {output['StandardErrorContent']}")
                        break
                except ssm_client.exceptions.InvocationDoesNotExist:
                    pass
                time.sleep(0.1)

            if output["Status"] != "Success":  # type: ignore
                log("Error: Command didn't finish successfully in time")
                return

            log(f"{time.ctime()} sm-connect-ssh-proxy: Starting SSH over SSM proxy")

            proxy_command = (
                "aws ssm start-session "
                + "--reason 'PackYak SSH' "
                + f"--region '{current_region}' "
                + f"--target '{instance_id}' "
                + "--document-name AWS-StartSSHSession "
                + "--parameters portNumber=%p"
            )

            # Start with the base SSH command
            command = ["ssh", "-4"]

            # Add each SSH option prefixed by '-o'
            for option in [
                f"User={user}",
                f"IdentityFile={ssh_key}",
                "IdentitiesOnly=yes",
                f"ProxyCommand={proxy_command}",
                "ServerAliveInterval=15",
                "ServerAliveCountMax=3",
                "PasswordAuthentication=no",
                "StrictHostKeyChecking=no",
                "UserKnownHostsFile=/dev/null",
            ]:
                command.extend(["-o", option])

            # Add each port forward prefixed by '-L'
            for port in port_forwards:
                command.extend(["-L", port])

            # Finally, add the instance ID
            command.append(instance_id)

            log(" ".join(command))  # Debugging: log the command

            # Execute the command
            process = subprocess.run(
                command,
                check=True,
            )
            return process.stdout, process.stderr

    asyncio.run(ssh(instance_id, user))


class Host:
    def __init__(self, name: str, props: dict[str, Any]):
        self.name = name
        self.props = props


async def parse_ssh_config(file_path: str = "~/.ssh/config") -> list[Host]:
    """Parse the SSH config file and return a list of Host objects with kwargs for each property."""
    hosts = []

    def parse_key_value(line: str) -> tuple[str, str] | None:
        split = re.split(r"\s+", line.strip(), maxsplit=1)
        if len(split) == 2:
            return split[0], split[1]
        else:
            return None

    async with aiofiles.open(file_path, "r") as f:
        lines = await f.readlines()

        queue_lines = collections.deque(lines)

        hosts: list[Host] = []

        while queue_lines:
            line = queue_lines.popleft()

            if line.startswith("Host "):
                parsed_host = parse_key_value(line)
                if parsed_host is None:
                    raise ValueError(
                        "Failed to parse host name from SSH config line:\n{line}"
                    )
                host = Host(name=parsed_host[1], props={})
                hosts.append(host)

                # now collect all of its key-value props
                while queue_lines:
                    # Peek at the next line without removing it from the queue
                    line = queue_lines[0]
                    if line.startswith("Host "):
                        # have found a new host, we're done
                        break
                    else:
                        key_value = parse_key_value(line)
                        if key_value is not None:
                            host.props[key_value[0]] = key_value[1]

    return hosts



import click
from packyak.cli.cli import cli
from packyak.util.emr import EMR, NodeType
import asyncio
import os


@cli.command()
@click.option(
    "--cluster-id",
    required=True,
    type=str,
    help="The unique identifier of the EMR cluster.",
)
@click.option(
    "--node-type",
    type=click.Choice(["primary", "core", "task"], case_sensitive=False),
    default=None,
    help="The node type to filter by.",
)
@click.option(
    "--profile",
    required=False,
    type=str,
    default=None,
    help="The AWS profile to use for the session.",
)
def instances(cluster_id: str, node_type: NodeType | None, profile: str | None):
    """
    Lists instances in an EMR cluster with optional filtering by node type.
    """
    if profile is not None:
        os.environ["AWS_PROFILE"] = profile
    emr = EMR()

    async def list_and_print_instances(
        node_type: NodeType, types_to_fetch: list[NodeType]
    ):
        output_messages = []
        for node_type in types_to_fetch:
            instances = await emr.list_instance_ids(cluster_id, node_type)
            if instances:
                output_messages.append(f"{node_type.capitalize()}:")
                for instance in instances:
                    output_messages.append(instance)

        for message in output_messages:
            click.echo(message)

    if node_type is None:
        types_to_fetch: list[NodeType] = ["primary", "core", "task"]
    else:
        types_to_fetch = [node_type]

    asyncio.run(
        list_and_print_instances(
            node_type if node_type is not None else "primary", types_to_fetch
        )
    )



from typing import Any
from dataclasses import dataclass
from enum import Enum
import gzip
from typing import List
from packyak.cli.node_type import NodeType

from aiobotocore.session import get_session

session = get_session()


class ClusterStatus(Enum):
    STARTING = "STARTING"
    BOOTSTRAPPING = "BOOTSTRAPPING"
    RUNNING = "RUNNING"
    WAITING = "WAITING"
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"
    TERMINATED_WITH_ERRORS = "TERMINATED_WITH_ERRORS"


@dataclass
class Cluster:
    cluster_id: str
    cluster_name: str
    cluster_status: ClusterStatus


class EMR:
    async def list_instance_ids(
        self, cluster_id: str, node_type: NodeType
    ) -> List[str]:
        """
        Get the instance IDs of nodes in an EMR cluster based on the node type.

        :param cluster_id: The unique identifier of the EMR cluster.
        :param node_type: The type of the node (primary, core, task).
        :return: A list of instance IDs.
        """
        instance_ids: list[str] = []
        async with session.create_client("emr") as emr:
            # Map NodeType to EMR instance group types
            instance_group_type = {
                "primary": "MASTER",
                "core": "CORE",
                "task": "TASK",
            }.get(node_type, None)

            if instance_group_type is None:
                raise ValueError(f"Invalid node type: {node_type}")

            response = await emr.list_instance_groups(ClusterId=cluster_id)
            for group in response["InstanceGroups"]:
                if group.get("InstanceGroupType") == instance_group_type:
                    instances_response = await emr.list_instances(
                        ClusterId=cluster_id,
                        InstanceGroupId=group["Id"],
                        InstanceStates=[
                            "AWAITING_FULFILLMENT",
                            "PROVISIONING",
                            "BOOTSTRAPPING",
                            "RUNNING",
                            "TERMINATED",
                        ],
                    )

                    for instance in instances_response["Instances"]:
                        instance_id = instance.get("Ec2InstanceId")
                        if instance_id:
                            instance_ids.append(instance_id)

        return instance_ids

    async def get_primary_node_instance_id(self, cluster_id: str) -> str:
        instance_id = await self.try_get_primary_node_instance_id(cluster_id)
        if instance_id is None:
            raise Exception(f"No primary instance found for EMR cluster {cluster_id}")
        return instance_id

    async def try_get_primary_node_instance_id(self, cluster_id: str) -> str | None:
        async with session.create_client("emr") as emr:
            response = await emr.list_instance_groups(ClusterId=cluster_id)
            for group in response["InstanceGroups"]:
                if "InstanceGroupType" not in group:
                    raise Exception("No instance group type")
                if group["InstanceGroupType"] == "MASTER":
                    if "Id" not in group:
                        raise Exception("No instance group ID")
                    instances = (
                        await emr.list_instances(
                            ClusterId=cluster_id,
                            InstanceGroupId=group["Id"],
                            InstanceStates=[
                                "AWAITING_FULFILLMENT",
                                "PROVISIONING",
                                "BOOTSTRAPPING",
                                "RUNNING",
                                "TERMINATED",
                            ],
                        )
                    )["Instances"]
                    if len(instances) > 0:
                        if "Ec2InstanceId" not in instances[0]:
                            raise Exception("No EC2 instance ID")
                        return instances[0]["Ec2InstanceId"]

        return None

    async def list_clusters(
        self,
        active_only: bool = True,
    ) -> List[Cluster]:
        """
        List EMR clusters based on their states asynchronously.

        :param active_only: If True, only active clusters are listed. Otherwise, all clusters are listed.
        :return: A list of Cluster objects representing the EMR clusters.
        """

        states = (
            [
                ClusterStatus.STARTING,
                ClusterStatus.BOOTSTRAPPING,
                ClusterStatus.RUNNING,
                ClusterStatus.WAITING,
            ]
            if active_only
            else [
                ClusterStatus.STARTING,
                ClusterStatus.BOOTSTRAPPING,
                ClusterStatus.RUNNING,
                ClusterStatus.WAITING,
                ClusterStatus.TERMINATING,
                ClusterStatus.TERMINATED,
                ClusterStatus.TERMINATED_WITH_ERRORS,
            ]
        )

        async with session.create_client("emr") as emr:
            response = await emr.list_clusters(
                ClusterStates=[state.value for state in states]
            )
            clusters = response.get("Clusters", [])

        def parse_cluster_status(cluster: Any) -> ClusterStatus:
            return ClusterStatus(cluster["Status"]["State"])

        def fail(msg: str):
            raise Exception(msg)

        return [
            Cluster(
                cluster_id=cluster.get("Id", fail("No cluster ID")),
                cluster_name=cluster.get("Name", fail("No cluster name")),
                cluster_status=parse_cluster_status(cluster),
            )
            for cluster in clusters
        ]

    _bucket_name_cache = {}

    async def get_cluster_log_uri(self, cluster_id: str) -> str:
        """
        Get the S3 location of the specified EMR cluster's logs asynchronously.
        This method caches the result to avoid repeated calls for the same cluster_id.

        :param cluster_id: The unique identifier of the EMR cluster.
        :return: The S3 URI where the cluster's logs are stored.
        """
        if cluster_id in self._bucket_name_cache:
            return self._bucket_name_cache[cluster_id]

        async with session.create_client("emr") as emr:
            try:
                cluster_description = await emr.describe_cluster(ClusterId=cluster_id)
                log_uri = cluster_description["Cluster"]["LogUri"]
                self._bucket_name_cache[cluster_id] = log_uri
                return log_uri
            except Exception as e:
                raise Exception(
                    f"Failed to get log URI for cluster {cluster_id}: {str(e)}"
                )

    async def get_bucket_name(self, cluster_id: str) -> str:
        log_uri = await self.get_cluster_log_uri(cluster_id)
        bucket_name = log_uri.split("/")[2]
        return bucket_name

    async def get_log_text(self, cluster_id: str, key: str) -> str:
        async with session.create_client("s3") as s3:
            response = await s3.get_object(
                Bucket=await self.get_bucket_name(cluster_id), Key=key
            )
            obj = await response["Body"].read()
            decompressed_data = gzip.decompress(obj)
            return decompressed_data.decode("utf-8")

    async def list_logs(self, cluster_id: str, instance_id: str) -> List[str]:
        """
        List all logs for a given EMR instance within a specified bucket.

        :param bucket_name: The name of the S3 bucket where logs are stored.
        :param instance_id: The unique identifier of the EMR instance.
        :return: A list of log file names.
        """
        log_files = []
        bucket_name = await self.get_bucket_name(cluster_id)
        async with session.create_client("s3") as s3:
            paginator = s3.get_paginator("list_objects_v2")
            async for page in paginator.paginate(
                Bucket=bucket_name, Prefix=f"{cluster_id}/node/{instance_id}/"
            ):
                for obj in page.get("Contents", []):
                    log_files.append(obj["Key"])
        return log_files
